/*
 * Copyright (C) 2007-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afs.c Paraslash's audio file selector. */

#include <signal.h>
#include <fnmatch.h>
#include "server.cmdline.h"
#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include <dirent.h> /* readdir() */
#include <sys/mman.h>
#include <sys/time.h>
#include "net.h"
#include "ipc.h"
#include "list.h"
#include "sched.h"
#include "signal.h"
#include "fd.h"

/** The osl tables used by afs. \sa blob.c. */
enum afs_table_num {
	/** Contains audio file information. See aft.c. */
	TBLNUM_AUDIO_FILES,
	/** The table for the paraslash attributes. See attribute.c. */
	TBLNUM_ATTRIBUTES,
	/**
	 * Paraslash's scoring system is based on Gaussian normal
	 * distributions, and the relevant data is stored in the rbtrees of an
	 * osl table containing only volatile columns.  See score.c for
	 * details.
	 */
	TBLNUM_SCORES,
	/**
	 * A standard blob table containing the mood definitions. For details
	 * see mood.c.
	 */
	TBLNUM_MOODS,
	/** A blob table containing lyrics on a per-song basis. */
	TBLNUM_LYRICS,
	/** Another blob table for images (for example album cover art). */
	TBLNUM_IMAGES,
	/** Yet another blob table for storing standard playlists. */
	TBLNUM_PLAYLIST,
	/** How many tables are in use? */
	NUM_AFS_TABLES
};

static struct afs_table afs_tables[NUM_AFS_TABLES] = {
	[TBLNUM_AUDIO_FILES] = {.init = aft_init},
	[TBLNUM_ATTRIBUTES] = {.init = attribute_init},
	[TBLNUM_SCORES] = {.init = score_init},
	[TBLNUM_MOODS] = {.init = moods_init},
	[TBLNUM_LYRICS] = {.init = lyrics_init},
	[TBLNUM_IMAGES] = {.init = images_init},
	[TBLNUM_PLAYLIST] = {.init = playlists_init},
};

struct command_task {
	/** The file descriptor for the local socket. */
	int fd;
	/**
	 * Value sent by the command handlers to identify themselves as
	 * children of the running para_server.
	 */
	uint32_t cookie;
	/** The associated task structure. */
	struct task task;
};

extern struct misc_meta_data *mmd;

static int server_socket;
static struct command_task command_task_struct;
static struct signal_task signal_task_struct;

static enum play_mode current_play_mode;
static char *current_mop; /* mode or playlist specifier. NULL means dummy mooe */


/**
 * A random number used to "authenticate" the connection.
 *
 * para_server picks this number by random before forking the afs process.  The
 * command handlers write this number together with the id of the shared memory
 * area containing the query. This way, a malicious local user has to know this
 * number to be able to cause the afs process to crash by sending fake queries.
 */
extern uint32_t afs_socket_cookie;

/**
 * Struct to let command handlers execute a callback in afs context.
 *
 * Commands that need to change the state of afs can't change the relevant data
 * structures directly because commands are executed in a child process, i.e.
 * they get their own virtual address space.
 *
 * This structure is used by \p send_callback_request() (executed from handler
 * context) in order to let the afs process call the specified function. An
 * instance of that structure is written to a shared memory area together with
 * the arguments to the callback function. The identifier of the shared memory
 * area is written to the command socket.
 *
 * The afs process accepts connections on the command socket and reads the
 * shared memory id, attaches the corresponing area, calls the given handler to
 * perform the desired action and to optionally compute a result.
 *
 * The result and a \p callback_result structure is then written to another
 * shared memory area. The identifier for that area is written to the handler's
 * command socket, so that the handler process can read the id, attach the
 * shared memory area and use the result.
 *
 * \sa struct callback_result.
 */
struct callback_query {
	/** The function to be called. */
	callback_function *handler;
	/** The number of bytes of the query */
	size_t query_size;
};

/**
 * Structure embedded in the result of a callback.
 *
 * If the callback produced a result, an instance of that structure is embeeded
 * into the shared memory area holding the result, mainly to let the command
 * handler know the size of the result.
 *
 * \sa struct callback_query.
 */
struct callback_result {
	/** The number of bytes of the result. */
	size_t result_size;
};

/**
 * Ask the afs process to call a given function.
 *
 * \param f The function to be called.
 * \param query Pointer to arbitrary data for the callback.
 * \param result Callback result will be stored here.
 *
 * This function creates a shared memory area, copies the buffer pointed to by
 * query to that area and notifies the afs process that \a f should be
 * called ASAP.
 *
 * \return Negative, on errors, the return value of the callback function
 * otherwise.
 *
 * \sa send_option_arg_callback_request(), send_standard_callback_request().
 */
int send_callback_request(callback_function *f, struct osl_object *query,
		struct osl_object *result)
{
	struct callback_query *cq;
	struct callback_result *cr;
	int ret, fd = -1, query_shmid, result_shmid;
	void *query_shm, *result_shm;
	char buf[sizeof(afs_socket_cookie) + sizeof(int)];
	size_t query_shm_size = sizeof(*cq);

	if (query)
		query_shm_size += query->size;
	ret = shm_new(query_shm_size);
	if (ret < 0)
		return ret;
	query_shmid = ret;
	ret = shm_attach(query_shmid, ATTACH_RW, &query_shm);
	if (ret < 0)
		goto out;
	cq = query_shm;
	cq->handler = f;
	cq->query_size = query_shm_size - sizeof(*cq);

	if (query)
		memcpy(query_shm + sizeof(*cq), query->data, query->size);
	ret = shm_detach(query_shm);
	if (ret < 0)
		goto out;

	*(uint32_t *) buf = afs_socket_cookie;
	*(int *) (buf + sizeof(afs_socket_cookie)) = query_shmid;

	ret = create_remote_socket(conf.afs_socket_arg);
	if (ret < 0)
		goto out;
	fd = ret;
	ret = send_bin_buffer(fd, buf, sizeof(buf));
	if (ret < 0)
		goto out;
	ret = recv_bin_buffer(fd, buf, sizeof(buf));
	if (ret < 0)
		goto out;
	if (ret != sizeof(int)) {
		ret = -E_AFS_SHORT_READ;
		goto out;
	}
	ret = *(int *) buf;
	if (ret <= 0)
		goto out;
	result_shmid = ret;
	ret = shm_attach(result_shmid, ATTACH_RO, &result_shm);
	if (ret >= 0) {
		assert(result);
		cr = result_shm;
		result->size = cr->result_size;
		result->data = para_malloc(result->size);
		memcpy(result->data, result_shm + sizeof(*cr), result->size);
		ret = shm_detach(result_shm);
		if (ret < 0)
			PARA_ERROR_LOG("can not detach result\n");
	} else
		PARA_ERROR_LOG("attach result failed: %d\n", ret);
	if (shm_destroy(result_shmid) < 0)
		PARA_ERROR_LOG("destroy result failed\n");
	ret = 1;
out:
	if (shm_destroy(query_shmid) < 0)
		PARA_ERROR_LOG("%s\n", "shm destroy error");
	if (fd >= 0)
		close(fd);
//	PARA_DEBUG_LOG("callback_ret: %d\n", ret);
	return ret;
}

/**
 * Send a callback request passing an options structure and an argument vector.
 *
 * \param options pointer to an arbitrary data structure.
 * \param argc Argument count.
 * \param argv Standard argument vector.
 * \param f The callback function.
 * \param result The result of the query is stored here.
 *
 * Some commands have a couple of options that are parsed in child context for
 * syntactic correctness and are stored in a special options structure for that
 * command. This function allows to pass such a structure together with a list
 * of further arguments (often a list of audio files) to the parent process.
 *
 * \sa send_standard_callback_request(), send_callback_request().
 */
int send_option_arg_callback_request(struct osl_object *options,
		int argc,  char * const * const argv, callback_function *f,
		struct osl_object *result)
{
	char *p;
	int i, ret;
	struct osl_object query = {.size = options? options->size : 0};

	for (i = 0; i < argc; i++)
		query.size += strlen(argv[i]) + 1;
	query.data = para_malloc(query.size);
	p = query.data;
	if (options) {
		memcpy(query.data, options->data, options->size);
		p += options->size;
	}
	for (i = 0; i < argc; i++) {
		strcpy(p, argv[i]); /* OK */
		p += strlen(argv[i]) + 1;
	}
	ret = send_callback_request(f, &query, result);
	free(query.data);
	return ret;
}

/**
 * Send a callback request with an argument vector only.
 *
 * \param argc The same meaning as in send_option_arg_callback_request().
 * \param argv The same meaning as in send_option_arg_callback_request().
 * \param f The same meaning as in send_option_arg_callback_request().
 * \param result The same meaning as in send_option_arg_callback_request().
 *
 * This is similar to send_option_arg_callback_request(), but no options buffer
 * is passed to the parent process.
 *
 * \return The return value of the underlying call to
 * send_option_arg_callback_request().
 */
int send_standard_callback_request(int argc,  char * const * const argv,
		callback_function *f, struct osl_object *result)
{
	return send_option_arg_callback_request(NULL, argc, argv, f, result);
}

static int action_if_pattern_matches(struct osl_row *row, void *data)
{
	struct pattern_match_data *pmd = data;
	struct osl_object name_obj;
	const char *p, *name;
	int ret = osl_get_object(pmd->table, row, pmd->match_col_num, &name_obj);
	const char *pattern_txt = (const char *)pmd->patterns.data;

	if (ret < 0)
		return ret;
	name = (char *)name_obj.data;
	if ((!name || !*name) && (pmd->pm_flags & PM_SKIP_EMPTY_NAME))
		return 1;
	if (!pmd->patterns.size && (pmd->pm_flags & PM_NO_PATTERN_MATCHES_EVERYTHING))
		return pmd->action(pmd->table, row, name, pmd->data);
	for (p = pattern_txt; p < pattern_txt + pmd->patterns.size;
			p += strlen(p) + 1) {
		ret = fnmatch(p, name, pmd->fnmatch_flags);
		if (ret == FNM_NOMATCH)
			continue;
		if (ret)
			return -E_FNMATCH;
		return pmd->action(pmd->table, row, name, pmd->data);
	}
	return 1;
}

/**
 * Execute the given function for each matching row.
 *
 * \param pmd Describes what to match and how.
 *
 * \return The return value of the underlying call to osl_rbtree_loop()
 * or osl_rbtree_loop_reverse().
 */
int for_each_matching_row(struct pattern_match_data *pmd)
{
	if (pmd->pm_flags & PM_REVERSE_LOOP)
		return osl_rbtree_loop_reverse(pmd->table, pmd->loop_col_num, pmd,
			action_if_pattern_matches);
	return osl_rbtree_loop(pmd->table, pmd->loop_col_num, pmd,
			action_if_pattern_matches);
}

/**
 * Compare two osl objects of string type.
 *
 * \param obj1 Pointer to the first object.
 * \param obj2 Pointer to the second object.
 *
 * In any case, only \p MIN(obj1->size, obj2->size) characters of each string
 * are taken into account.
 *
 * \return It returns an integer less than, equal to, or greater than zero if
 * \a obj1 is found, respectively, to be less than, to match, or be greater than
 * obj2.
 *
 * \sa strcmp(3), strncmp(3), osl_compare_func.
 */
int string_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	const char *str1 = (const char *)obj1->data;
	const char *str2 = (const char *)obj2->data;
	return strncmp(str1, str2, PARA_MIN(obj1->size, obj2->size));
}

/*
 * write input from fd to dynamically allocated buffer,
 * but maximal max_size byte.
 */
static int fd2buf(int fd, unsigned max_size, struct osl_object *obj)
{
	const size_t chunk_size = 1024;
	size_t size = 2048, received = 0;
	int ret;
	char *buf = para_malloc(size);

	for (;;) {
		ret = recv_bin_buffer(fd, buf + received, chunk_size);
		if (ret <= 0)
			break;
		received += ret;
		if (received + chunk_size >= size) {
			size *= 2;
			ret = -E_INPUT_TOO_LARGE;
			if (size > max_size)
				break;
			buf = para_realloc(buf, size);
		}
	}
	obj->data = buf;
	obj->size = received;
	if (ret < 0)
		free(buf);
	return ret;
}

/**
 * Read data from a file descriptor, and send it to the afs process.
 *
 * \param fd File descriptor to read data from.
 * \param arg_obj Pointer to the arguments to \a f.
 * \param f The callback function.
 * \param max_len Don't read more than that many bytes from stdin.
 * \param result The result of the query is stored here.
 *
 * This function is used by commands that wish to let para_server store
 * arbitrary data specified by the user (for instance the add_blob family of
 * commands). First, at most \a max_len bytes are read from \a fd, the result
 * is concatenated with the buffer given by \a arg_obj, and the combined buffer
 * is made available to the parent process via shared memory.
 *
 * \return Negative on errors, the return value of the underlying call to
 * send_callback_request() otherwise.
 */
int stdin_command(int fd, struct osl_object *arg_obj, callback_function *f,
		unsigned max_len, struct osl_object *result)
{
	struct osl_object query, stdin_obj;
	int ret;

	ret = send_buffer(fd, AWAITING_DATA_MSG);
	if (ret < 0)
		return ret;
	ret = fd2buf(fd, max_len, &stdin_obj);
	if (ret < 0)
		return ret;
	query.size = arg_obj->size + stdin_obj.size;
	query.data = para_malloc(query.size);
	memcpy(query.data, arg_obj->data, arg_obj->size);
	memcpy((char *)query.data + arg_obj->size, stdin_obj.data, stdin_obj.size);
	free(stdin_obj.data);
	ret = send_callback_request(f, &query, result);
	free(query.data);
	return ret;
}

static int pass_afd(int fd, char *buf, size_t size)
{
	struct msghdr msg = {.msg_iov = NULL};
	struct cmsghdr *cmsg;
	char control[255];
	int ret;
	struct iovec iov;

	iov.iov_base = buf;
	iov.iov_len  = size;

	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);

	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_RIGHTS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(int));
	*(int *)CMSG_DATA(cmsg) = fd;

	/* Sum of the length of all control messages in the buffer */
	msg.msg_controllen = cmsg->cmsg_len;
	PARA_DEBUG_LOG("passing %zu bytes and fd %d\n", size, fd);
	ret = sendmsg(server_socket, &msg, 0);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		return ret;
	}
	return 1;
}

/**
 * Open the audio file with highest score.
 *
 * This stores all information for streaming the "best" audio file in a shared
 * memory area. The id of that area and an open file descriptor for the next
 * audio file are passed to the server process.
 *
 * \return Standard.
 *
 * \sa open_and_update_audio_file().
 */
int open_next_audio_file(void)
{
	struct osl_row *aft_row;
	struct audio_file_data afd;
	int ret, shmid;
	char buf[8];
	long score;
again:
	PARA_NOTICE_LOG("getting next audio file\n");
	ret = score_get_best(&aft_row, &score);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		goto no_admissible_files;
	}
	ret = open_and_update_audio_file(aft_row, score, &afd);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		ret = score_delete(aft_row);
		if (ret < 0) {
			PARA_ERROR_LOG("%s\n", para_strerror(-ret));
			goto no_admissible_files;
		}
		goto again;
	}
	shmid = ret;
	if (!write_ok(server_socket)) {
		ret = -E_AFS_SOCKET;
		goto destroy;
	}
	*(uint32_t *)buf = NEXT_AUDIO_FILE;
	*(uint32_t *)(buf + 4) = (uint32_t)shmid;
	ret = pass_afd(afd.fd, buf, 8);
	close(afd.fd);
	if (ret >= 0)
		return ret;
destroy:
	shm_destroy(shmid);
	return ret;
no_admissible_files:
	*(uint32_t *)buf = NO_ADMISSIBLE_FILES;
	*(uint32_t *)(buf + 4) = (uint32_t)0;
	return send_bin_buffer(server_socket, buf, 8);
}

/* Never fails if arg == NULL */
static int activate_mood_or_playlist(char *arg, int *num_admissible)
{
	enum play_mode mode;
	int ret;

	if (!arg) {
		ret = change_current_mood(NULL); /* always successful */
		mode = PLAY_MODE_MOOD;
	} else {
		if (!strncmp(arg, "p/", 2)) {
			ret = playlist_open(arg + 2);
			mode = PLAY_MODE_PLAYLIST;
		} else if (!strncmp(arg, "m/", 2)) {
			ret = change_current_mood(arg + 2);
			mode = PLAY_MODE_MOOD;
		} else
			return -E_AFS_SYNTAX;
		if (ret < 0)
			return ret;
	}
	if (num_admissible)
		*num_admissible = ret;
	current_play_mode = mode;
	if (arg != current_mop) {
		free(current_mop);
		if (arg) {
			current_mop = para_strdup(arg);
			mmd_lock();
			strncpy(mmd->afs_mode_string, arg,
				sizeof(mmd->afs_mode_string));
			mmd->afs_mode_string[sizeof(mmd->afs_mode_string) - 1] = '\0';
			mmd_unlock();
		} else {
			mmd_lock();
			strcpy(mmd->afs_mode_string, "dummy");
			mmd_unlock();
			current_mop = NULL;
		}
	}
	return 1;
}

static int com_select_callback(const struct osl_object *query,
		struct osl_object *result)
{
	struct para_buffer pb = {.buf = NULL};
	char *arg = query->data;
	int num_admissible, ret;

	ret = clear_score_table();
	if (ret < 0)
		return ret;
	if (current_play_mode == PLAY_MODE_MOOD)
		close_current_mood();
	else
		playlist_close();
	ret = activate_mood_or_playlist(arg, &num_admissible);
	if (ret < 0) {
		para_printf(&pb, "%s\n", para_strerror(-ret));
		para_printf(&pb, "switching back to %s\n", current_mop?
			current_mop : "dummy");
		ret = activate_mood_or_playlist(current_mop, &num_admissible);
		if (ret < 0) {
			para_printf(&pb, "failed, switching to dummy\n");
			activate_mood_or_playlist(NULL, &num_admissible);
		}
	}
	para_printf(&pb, "activated %s (%d admissible files)\n", current_mop?
		current_mop : "dummy mood", num_admissible);
	result->data = pb.buf;
	result->size = pb.size;
	return 1;
}

int com_select(int fd, int argc, char * const * const argv)
{
	int ret;
	struct osl_object query, result;

	if (argc != 2)
		return -E_AFS_SYNTAX;
	query.data = argv[1];
	query.size = strlen(argv[1]) + 1;
	ret = send_callback_request(com_select_callback, &query,
		&result);
	if (ret > 0 && result.data && result.size) {
		ret = send_va_buffer(fd, "%s", (char *)result.data);
		free(result.data);
	}
	return ret;
}

static void init_admissible_files(char *arg)
{
	if (activate_mood_or_playlist(arg, NULL) < 0)
		activate_mood_or_playlist(NULL, NULL); /* always successful */
}

static int setup_command_socket_or_die(void)
{
	int ret, socket_fd;
	char *socket_name = conf.afs_socket_arg;
	struct sockaddr_un unix_addr;

	unlink(socket_name);
	ret = create_local_socket(socket_name, &unix_addr,
		S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (ret < 0) {
		PARA_EMERG_LOG("%s: %s\n", para_strerror(-ret), socket_name);
		exit(EXIT_FAILURE);
	}
	socket_fd = ret;
	if (listen(socket_fd , 5) < 0) {
		PARA_EMERG_LOG("can not listen on socket\n");
		exit(EXIT_FAILURE);
	}
	ret = mark_fd_nonblocking(socket_fd);
	if (ret < 0) {
		close(socket_fd);
		return ret;
	}
	PARA_INFO_LOG("listening on socket %s (fd %d)\n", socket_name, ret);
	return socket_fd;
}

static void close_afs_tables(void)
{
	int i;
	PARA_NOTICE_LOG("closing afs_tables\n");
	for (i = 0; i < NUM_AFS_TABLES; i++)
		afs_tables[i].close();
}

static char *database_dir;

static void get_database_dir(void)
{
	if (!database_dir) {
		if (conf.afs_database_dir_given)
			database_dir = para_strdup(conf.afs_database_dir_arg);
		else {
			char *home = para_homedir();
			database_dir = make_message(
				"%s/.paraslash/afs_database", home);
			free(home);
		}
	}
	PARA_INFO_LOG("afs_database dir %s\n", database_dir);
}

static int make_database_dir(void)
{
	int ret;

	get_database_dir();
	ret = para_mkdir(database_dir, 0777);
	if (ret >= 0 || is_errno(-ret, EEXIST))
		return 1;
	return ret;
}

static int open_afs_tables(void)
{
	int i, ret;

	get_database_dir();
	PARA_NOTICE_LOG("opening %u osl tables in %s\n", NUM_AFS_TABLES,
		database_dir);
	for (i = 0; i < NUM_AFS_TABLES; i++) {
		ret = afs_tables[i].open(database_dir);
		if (ret >= 0)
			continue;
		PARA_ERROR_LOG("%s init: %s\n", afs_tables[i].name,
			para_strerror(-ret));
		break;
	}
	if (ret >= 0)
		return ret;
	while (i)
		afs_tables[--i].close();
	return ret;
}

static void unregister_tasks(void)
{
	unregister_task(&command_task_struct.task);
	unregister_task(&signal_task_struct.task);
}

static void signal_pre_select(struct sched *s, struct task *t)
{
	struct signal_task *st = t->private_data;
	t->ret = 1;
	para_fd_set(st->fd, &s->rfds, &s->max_fileno);
}

static void signal_post_select(struct sched *s, struct task *t)
{
	struct signal_task *st = t->private_data;
	t->ret = -E_AFS_PARENT_DIED;
	if (getppid() == 1)
		goto err;
	t->ret = 1;
	if (!FD_ISSET(st->fd, &s->rfds))
		return;
	st->signum = para_next_signal();
	t->ret = 1;
	if (st->signum == SIGUSR1)
		return; /* ignore SIGUSR1 */
	if (st->signum == SIGHUP) {
		close_afs_tables();
		t->ret = open_afs_tables();
		if (t->ret < 0)
			goto err;
		init_admissible_files(current_mop);
		return;
	}
	t->ret = -E_AFS_SIGNAL;
err:
	PARA_NOTICE_LOG("%s\n", para_strerror(-t->ret));
	unregister_tasks();
}

static void register_signal_task(void)
{
	struct signal_task *st = &signal_task_struct;
	st->fd = para_signal_init();
	PARA_INFO_LOG("signal pipe: fd %d\n", st->fd);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGPIPE);
	para_install_sighandler(SIGHUP);

	st->task.pre_select = signal_pre_select;
	st->task.post_select = signal_post_select;
	st->task.private_data = st;
	sprintf(st->task.status, "signal task");
	register_task(&st->task);
}

static struct list_head afs_client_list;

/** Describes on connected afs client. */
struct afs_client {
	/** Position in the afs client list. */
	struct list_head node;
	/** The socket file descriptor for this client. */
	int fd;
	/** The time the client connected. */
	struct timeval connect_time;
};

static void command_pre_select(struct sched *s, struct task *t)
{
	struct command_task *ct = t->private_data;
	struct afs_client *client;

	para_fd_set(server_socket, &s->rfds, &s->max_fileno);
	para_fd_set(ct->fd, &s->rfds, &s->max_fileno);
	list_for_each_entry(client, &afs_client_list, node)
		para_fd_set(client->fd, &s->rfds, &s->max_fileno);
	t->ret = 1;
}

/*
 * On errors, negative value is written to fd.
 * On success: If query produced a result, the result_shmid is written to fd.
 * Otherwise, zero is written.
 */
static int call_callback(int fd, int query_shmid)
{
	void *query_shm, *result_shm;
	struct callback_query *cq;
	struct callback_result *cr;
	struct osl_object query, result = {.data = NULL};
	int result_shmid = -1, ret, ret2;

	ret = shm_attach(query_shmid, ATTACH_RW, &query_shm);
	if (ret < 0)
		goto out;
	cq = query_shm;
	query.data = (char *)query_shm + sizeof(*cq);
	query.size = cq->query_size;
	ret = cq->handler(&query, &result);
	ret2 = shm_detach(query_shm);
	if (ret2 < 0 && ret >= 0)
		ret = ret2;
	if (ret < 0)
		goto out;
	ret = 0;
	if (!result.data || !result.size)
		goto out;
	ret = shm_new(result.size + sizeof(struct callback_result));
	if (ret < 0)
		goto out;
	result_shmid = ret;
	ret = shm_attach(result_shmid, ATTACH_RW, &result_shm);
	if (ret < 0)
		goto out;
	cr = result_shm;
	cr->result_size = result.size;
	memcpy(result_shm + sizeof(*cr), result.data, result.size);
	ret = shm_detach(result_shm);
	if (ret < 0)
		goto out;
	ret = result_shmid;
out:
	free(result.data);
	ret2 = send_bin_buffer(fd, (char *)&ret, sizeof(int));
	if (ret < 0 || ret2 < 0) {
		if (result_shmid >= 0)
			if (shm_destroy(result_shmid) < 0)
				PARA_ERROR_LOG("destroy result failed\n");
		if (ret >= 0)
			ret = ret2;
	}
	return ret;
}

static void execute_server_command(void)
{
	char buf[8];
	int ret = recv_bin_buffer(server_socket, buf, sizeof(buf) - 1);

	if (ret <= 0) {
		if (ret < 0)
			PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		return;
	}
	buf[ret] = '\0';
	PARA_DEBUG_LOG("received: %s\n", buf);
	if (!strcmp(buf, "new")) {
		ret = open_next_audio_file();
		if (ret < 0) {
			PARA_EMERG_LOG("%s\n", para_strerror(-ret));
			unregister_tasks();
		}
		return;
	}
	PARA_ERROR_LOG("unknown command\n");

}

static void execute_afs_command(int fd, uint32_t expected_cookie)
{
	uint32_t cookie;
	int query_shmid;
	char buf[sizeof(cookie) + sizeof(query_shmid)];
	int ret = recv_bin_buffer(fd, buf, sizeof(buf));

	if (ret < 0) {
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
		return;
	}
	if (ret != sizeof(buf)) {
		PARA_NOTICE_LOG("short read (%d bytes, expected %lu)\n",
			ret, (long unsigned) sizeof(buf));
		return;
	}
	cookie = *(uint32_t *)buf;
	if (cookie != expected_cookie) {
		PARA_NOTICE_LOG("received invalid cookie(got %u, expected %u)\n",
			(unsigned)cookie, (unsigned)expected_cookie);
		return;
	}
	query_shmid = *(int *)(buf + sizeof(cookie));
	if (query_shmid < 0) {
		PARA_WARNING_LOG("received invalid query shmid %d)\n",
			query_shmid);
		return;
	}
	/* Ignore return value: Errors might be OK here. */
	call_callback(fd, query_shmid);
}

/** Shutdown connection if query has not arrived until this many seconds. */
#define AFS_CLIENT_TIMEOUT 3

static void command_post_select(struct sched *s, struct task *t)
{
	struct command_task *ct = t->private_data;
	struct sockaddr_un unix_addr;
	struct afs_client *client, *tmp;
	int fd;
	if (FD_ISSET(server_socket, &s->rfds))
		execute_server_command();

	/* Check the list of connected clients. */
	list_for_each_entry_safe(client, tmp, &afs_client_list, node) {
		if (FD_ISSET(client->fd, &s->rfds))
			execute_afs_command(client->fd, ct->cookie);
		else { /* prevent bogus connection flooding */
			struct timeval diff;
			tv_diff(now, &client->connect_time, &diff);
			if (diff.tv_sec < AFS_CLIENT_TIMEOUT)
				continue;
			PARA_WARNING_LOG("connection timeout\n");
		}
		close(client->fd);
		list_del(&client->node);
		free(client);
	}
	/* Accept connections on the local socket. */
	if (!FD_ISSET(ct->fd, &s->rfds))
		goto out;
	t->ret = para_accept(ct->fd, &unix_addr, sizeof(unix_addr));
	if (t->ret < 0) {
		PARA_NOTICE_LOG("%s\n", para_strerror(-t->ret));
		goto out;
	}
	fd = t->ret;
	t->ret = mark_fd_nonblocking(fd);
	if (t->ret < 0) {
		PARA_NOTICE_LOG("%s\n", para_strerror(-t->ret));
		close(fd);
		goto out;
	}
	client = para_malloc(sizeof(*client));
	client->fd = fd;
	client->connect_time = *now;
	para_list_add(&client->node, &afs_client_list);
out:
	t->ret = 1;
}

static void register_command_task(uint32_t cookie)
{
	struct command_task *ct = &command_task_struct;
	ct->fd = setup_command_socket_or_die();
	ct->cookie = cookie;

	ct->task.pre_select = command_pre_select;
	ct->task.post_select = command_post_select;
	ct->task.private_data = ct;
	sprintf(ct->task.status, "command task");
	register_task(&ct->task);
}

static void register_tasks(uint32_t cookie)
{
	register_signal_task();
	register_command_task(cookie);
}

/**
 * Initialize the audio file selector process.
 *
 * \param cookie The value used for "authentication".
 * \param socket_fd File descriptor used for communication with the server.
 */
__noreturn void afs_init(uint32_t cookie, int socket_fd)
{
	struct sched s;
	int i, ret;

	INIT_LIST_HEAD(&afs_client_list);
	for (i = 0; i < NUM_AFS_TABLES; i++)
		afs_tables[i].init(&afs_tables[i]);
	ret = open_afs_tables();

	if (ret < 0) {
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	server_socket = socket_fd;
	ret = mark_fd_nonblocking(server_socket);
	if (ret < 0)
		exit(EXIT_FAILURE);
	PARA_INFO_LOG("server_socket: %d, afs_socket_cookie: %u\n",
		server_socket, (unsigned) cookie);
	init_admissible_files(conf.afs_initial_mode_arg);
	register_tasks(cookie);
	s.default_timeout.tv_sec = 0;
	s.default_timeout.tv_usec = 999 * 1000;
	ret = schedule(&s);
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	close_afs_tables();
	exit(EXIT_FAILURE);
}

static int create_tables_callback(const struct osl_object *query,
		__a_unused struct osl_object *result)
{
	uint32_t table_mask = *(uint32_t *)query->data;
	int i, ret;

	close_afs_tables();
	for (i = 0; i < NUM_AFS_TABLES; i++) {
		struct afs_table *t = &afs_tables[i];

		if (!(table_mask & (1 << i)))
			continue;
		if (!t->create)
			continue;
		ret = t->create(database_dir);
		if (ret < 0)
			return ret;
	}
	ret = open_afs_tables();
	return ret < 0? ret: 0;
}

int com_init(int fd, int argc, char * const * const argv)
{
	int i, j, ret;
	uint32_t table_mask = (1 << (NUM_AFS_TABLES + 1)) - 1;
	struct osl_object query = {.data = &table_mask,
		.size = sizeof(table_mask)};

	ret = make_database_dir();
	if (ret < 0)
		return ret;
	if (argc != 1) {
		table_mask = 0;
		for (i = 1; i < argc; i++) {
			for (j = 0; j < NUM_AFS_TABLES; j++) {
				struct afs_table *t = &afs_tables[j];

				if (strcmp(argv[i], t->name))
					continue;
				table_mask |= (1 << j);
				break;
			}
			if (j == NUM_AFS_TABLES)
				return -E_BAD_TABLE_NAME;
		}
	}
	ret = send_callback_request(create_tables_callback, &query, NULL);
	if (ret < 0)
		return ret;
	return send_va_buffer(fd, "successfully created afs table(s)\n");
}

/**
 * Flags for the check command.
 *
 * \sa com_check().
 */
enum com_check_flags {
	/** Check the audio file table. */
	CHECK_AFT = 1,
	/** Check the mood table. */
	CHECK_MOODS = 2,
	/** Check the playlist table. */
	CHECK_PLAYLISTS = 4
};

int com_check(int fd, int argc, char * const * const argv)
{
	unsigned flags = 0;
	int i, ret;
	struct osl_object result;

	for (i = 1; i < argc; i++) {
		const char *arg = argv[i];
		if (arg[0] != '-')
			break;
		if (!strcmp(arg, "--")) {
			i++;
			break;
		}
		if (!strcmp(arg, "-a")) {
			flags |= CHECK_AFT;
			continue;
		}
		if (!strcmp(arg, "-p")) {
			flags |= CHECK_PLAYLISTS;
			continue;
		}
		if (!strcmp(arg, "-m")) {
			flags |= CHECK_MOODS;
			continue;
		}
		return -E_AFS_SYNTAX;
	}
	if (i < argc)
		return -E_AFS_SYNTAX;
	if (!flags)
		flags = ~0U;
	if (flags & CHECK_AFT) {
		ret = send_callback_request(aft_check_callback, NULL, &result);
		if (ret < 0)
			return ret;
		if (ret > 0) {
			ret = send_buffer(fd, (char *) result.data);
			free(result.data);
			if (ret < 0)
				return ret;
		}
	}
	if (flags & CHECK_PLAYLISTS) {
		ret = send_callback_request(playlist_check_callback, NULL, &result);
		if (ret < 0)
			return ret;
		if (ret > 0) {
			ret = send_buffer(fd, (char *) result.data);
			free(result.data);
			if (ret < 0)
				return ret;
		}
	}
	if (flags & CHECK_MOODS) {
		ret = send_callback_request(mood_check_callback, NULL, &result);
		if (ret < 0)
			return ret;
		if (ret > 0) {
			ret = send_buffer(fd, (char *) result.data);
			free(result.data);
			if (ret < 0)
				return ret;
		}
	}
	return 1;
}

void afs_event(enum afs_events event, struct para_buffer *pb,
		void *data)
{
	int i, ret;

	for (i = 0; i < NUM_AFS_TABLES; i++) {
		struct afs_table *t = &afs_tables[i];
		if (!t->event_handler)
			continue;
		ret = t->event_handler(event, pb, data);
		if (ret < 0)
			PARA_CRIT_LOG("%s\n", para_strerror(-ret));
	}
}

int images_event_handler(__a_unused enum afs_events event,
	__a_unused  struct para_buffer *pb, __a_unused void *data)
{
	return 1;
}

int lyrics_event_handler(__a_unused enum afs_events event,
	__a_unused struct para_buffer *pb, __a_unused void *data)
{
	return 1;
}
