/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file close_on_fork.c manage a list of fds that should be closed on fork */
#include "para.h"
#include "list.h"
#include "string.h"

static struct list_head close_on_fork_list;
static int initialized;

/**
 * describes an element of the close-on-fork list
 *
 * \sa list.h
 */
struct close_on_fork {
	/** the file descriptor which should be closed after fork() */
	int fd;
	/** the position in the close-on-fork list */
	struct list_head node;
};

/**
 * add one file descriptor to the close-on-fork list
 *
 * \param fd the file descriptor to add
 */
void add_close_on_fork_list(int fd)
{
	struct close_on_fork *cof = para_malloc(sizeof(struct close_on_fork));

	if (!initialized) {
		INIT_LIST_HEAD(&close_on_fork_list);
		initialized = 1;
	}
	cof->fd = fd;
	para_list_add(&cof->node, &close_on_fork_list);
}


/**
 * delete one file descriptor from the close-on-fork list
 *
 * \param fd the file descriptor to delete
 *
 * Noop if \a fd does not belong to the close-on-fork list.
 */
void del_close_on_fork_list(int fd)
{
	struct close_on_fork *cof, *tmp;

	if (!initialized)
		return;
	list_for_each_entry_safe(cof, tmp, &close_on_fork_list, node) {
		if (fd != cof->fd)
			continue;
		list_del(&cof->node);
		free(cof);
	}
}

/**
 * call close(3) for each fd in the close-on-fork list
 */
void close_listed_fds(void)
{
	struct close_on_fork *cof;

	if (!initialized)
		return;
	list_for_each_entry(cof, &close_on_fork_list, node) {
		PARA_DEBUG_LOG("closing fd %d\n", cof->fd);
		close(cof->fd);
	}
}
