/*
 * Copyright (C) 1998-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file fade.c a volume fader and alarm clock */

#include <sys/types.h>
#include <dirent.h>

#include "fade.cmdline.h"
#include "para.h"
#include "fd.h"

#include <stropts.h>
#include <ctype.h>
#include <stdlib.h> /* EXIT_SUCCESS */
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <linux/soundcard.h>
#include "string.h"


struct fade_args_info conf;

void para_log(__a_unused int ll, const char *fmt,...)
{
	va_list argp;
	time_t t1;
	struct tm *tm;

	time(&t1);
	tm = localtime(&t1);
	printf("%d:%02d:%02d ", tm->tm_hour, tm->tm_min, tm->tm_sec);
	va_start(argp, fmt);
	vprintf(fmt, argp);
	va_end(argp);
}

/*
 * open mixer device
 */
static int open_mixer(void)
{
	return open(conf.mixer_device_arg, O_RDWR, 0);
}

/*
 * get volume via mixer_fd
 */
static int do_get_vol(int mixer_fd)
{
	int volume;

	if (ioctl(mixer_fd, MIXER_READ(SOUND_MIXER_VOLUME), &volume) < 0)
		return -1;
	/* take the mean value of left and right volume */
	return (volume % 256 + (volume >> 8)) / 2;
}

/*
 * open mixer, get volume and close mixer
 */
static int get_vol(void)
{
	int mixer_fd;
	int volume;

	mixer_fd = open_mixer();
	if (mixer_fd < 0)
		return -1;
	volume = do_get_vol(mixer_fd);
	close(mixer_fd);
	return volume;
}

/*
 * set volume via mixer_fd
 */
static int do_set_vol(int mixer_fd, int volume)
{
	int tmp = (volume << 8) + volume;

	if (ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_VOLUME), &tmp) < 0)
		return -1;
	return 1;
}

/*
 * open mixer, set volume and close mixer
 */
static int set_vol(int volume)
{
	int mixer_fd, ret = open_mixer();

	if (ret < 0)
		return ret;
	mixer_fd = ret;
	ret = do_set_vol(mixer_fd, volume);
	close(mixer_fd);
	return ret;
}

/*
 * Open mixer, get volume, fade to new_vol in secs seconds and
 * close mixer
 */
static void fade(int new_vol, int fade_time)
{
	int vol, mixer_fd, diff, incr;
	unsigned secs;
	struct timespec ts;
	unsigned long long tmp, tmp2; /* Careful with that axe, Eugene! */

	if (fade_time <= 0)
		return;
	secs = fade_time;
	PARA_NOTICE_LOG("fading to %d in %d seconds\n", new_vol, secs);
	mixer_fd = open_mixer();
	if (mixer_fd < 0)
		return;
	vol = do_get_vol(mixer_fd);
	if (vol < 0)
		goto out;
	diff = new_vol - vol;
	if (!diff) {
		sleep(secs);
		goto out;
	}
	incr = diff > 0? 1: -1;
	diff = diff > 0? diff: -diff;
	tmp = secs * 1000 / diff;
	tmp2 = tmp % 1000;
	while ((new_vol - vol) * incr > 0) {
		ts.tv_nsec = tmp2 * 1000000; /* really nec ?*/
		ts.tv_sec = tmp / 1000; /* really nec ?*/
		//printf("ts.tv_sec: %i\n", ts.tv_nsec);
		vol += incr;
		if (do_set_vol(mixer_fd, vol) < 0)
			goto out;
		//printf("vol = %i\n", vol);
		nanosleep(&ts, NULL);
	}
out:
	close(mixer_fd);
}

static void client_cmd(const char *cmd)
{
	int ret, fds[3] = {0, 0, 0};
	pid_t pid;
	char *cmdline = make_message(BINDIR "/para_client %s", cmd);

	PARA_INFO_LOG("%s\n", cmdline);
	ret = para_exec_cmdline_pid(&pid, cmdline, fds);
	free(cmdline);
	if (ret < 0)
		exit(EXIT_FAILURE);
	do
		ret = wait(NULL);
	while (ret != -1 && errno != ECHILD);
}

static void change_afs_mode_and_play(char *afs_mode)
{
	char *cmd;

	client_cmd("stop");
	if (!afs_mode)
		return;
	cmd = make_message("select %s\n", afs_mode);
	client_cmd(cmd);
	free(cmd);
	client_cmd("play");
}

/*
 * sleep
 */
static void sweet_dreams(void)
{
	time_t t1, wake_time_epoch;
	unsigned int delay;
	struct tm *tm;
	int min = conf.wake_min_arg;
	char *fa_mode = conf.fa_mode_arg;
	char *wake_mode = conf.wake_mode_arg;
	char *sleep_mode = conf.sleep_mode_arg;
	int wf = conf.wake_fade_arg;
	int sf = conf.fa_fade_arg;
	int wv = conf.wake_vol_arg;
	int sv = conf.fa_vol_arg;
	int iv = conf.sleep_ivol_arg;

	/* calculate wake time */
	time(&t1);
	if (conf.wake_hour_given) {
		int hour = conf.wake_hour_arg;
		tm = localtime(&t1);
		if (tm->tm_hour > hour || (tm->tm_hour == hour && tm->tm_min> min)) {
			t1 += 86400; /* wake time is tomorrow */
			tm = localtime(&t1);
		}
		tm->tm_hour = hour;
		tm->tm_min = min;
		tm->tm_sec = 0;
	} else {
		t1 += 9 * 60 * 60; /* nine hours from now */
		PARA_INFO_LOG("default wake time: %lu\n", t1);
		tm = localtime(&t1);
	}
	wake_time_epoch = mktime(tm);
	PARA_INFO_LOG("waketime: %s", asctime(tm));
	client_cmd("stop");
	sleep(1);
	if (sf) {
		PARA_INFO_LOG("initial volume: %d\n", iv);
		set_vol(iv);
		change_afs_mode_and_play(fa_mode);
		fade(sv, sf);
	} else
		set_vol(sf);
	change_afs_mode_and_play(sleep_mode);
	if (!wf)
		return;
	for (;;) {
		time(&t1);
		if (wake_time_epoch <= t1 + wf)
			break;
		delay = wake_time_epoch - t1 - wf;
		PARA_INFO_LOG("sleeping %u seconds (%u:%02u)\n",
			delay, delay / 3600,
			(delay % 3600) / 60);
		sleep(delay);
	}
	change_afs_mode_and_play(wake_mode);
	fade(wv, wf);
	PARA_INFO_LOG("fade complete, returning\n");
}

static void snooze(void)
{
	unsigned sleep_time;

	if (conf.snooze_time_arg <= 0)
		return;
	sleep_time = conf.snooze_time_arg;
	if (get_vol() < conf.snooze_out_vol_arg)
		set_vol(conf.snooze_out_vol_arg);
	else
		fade(conf.snooze_out_vol_arg, conf.snooze_out_fade_arg);
	client_cmd("pause");
	PARA_NOTICE_LOG("%d seconds snooze time...\n", conf.snooze_time_arg);
	sleep(sleep_time);
	client_cmd("play");
	fade(conf.snooze_in_vol_arg, conf.snooze_in_fade_arg);
}

static int configfile_exists(void)
{
	static char *config_file;

	if (!conf.config_file_given) {
		char *home = para_homedir();
		free(config_file);
		config_file = make_message("%s/.paraslash/fade.conf", home);
		free(home);
		conf.config_file_arg = config_file;
	}
	return file_exists(conf.config_file_arg);
}


int main(int argc, char *argv[])
{
	int ret;

	if (fade_cmdline_parser(argc, argv, &conf))
		exit(EXIT_FAILURE);
	HANDLE_VERSION_FLAG("fade", conf);
	ret = configfile_exists();
	if (!ret && conf.config_file_given) {
		PARA_EMERG_LOG("can not read config file %s\n",
			conf.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (ret) {
		struct fade_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0
		};
		fade_cmdline_parser_config_file(conf.config_file_arg,
			&conf, &params);
	}
	ret = open_mixer();
	if (ret < 0) {
		PARA_EMERG_LOG("can not open mixer device %s.",
			conf.mixer_device_arg);
		exit(EXIT_FAILURE);
	}
	close(ret);
	ret = 0;
//	setlinebuf(stdout);
	if (!strcmp(conf.mode_arg, "sleep")) {
		sweet_dreams();
		goto out;
	}
	if (!strcmp(conf.mode_arg, "fade")) {
		fade(conf.fade_vol_arg, conf.fade_time_arg);
		goto out;
	}
	if (!strcmp(conf.mode_arg, "snooze")) {
		snooze();
		goto out;
	}
	ret = -1;
out:
	return ret < 0? EXIT_FAILURE : EXIT_SUCCESS;
}
